package java.lang.invoke;

import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Retention;
import java.lang.annotation.ElementType;
import java.lang.annotation.Target;
import java.util.Arrays;
import java.util.List;

public abstract class MethodHandle
{
    private final MethodType type;
    final LambdaForm form;
    MethodHandle asTypeCache;
    byte customizationCount;
    private static final long FORM_OFFSET;
    
    public MethodType type() {
        return this.type;
    }
    
    MethodHandle(MethodType type, LambdaForm lambdaForm) {
        type.getClass();
        lambdaForm.getClass();
        this.type = type;
        (this.form = lambdaForm.uncustomize()).prepare();
    }
    
    @MethodHandle.PolymorphicSignature
    public final native Object invokeExact(Object... p0) throws Throwable;
    
    @MethodHandle.PolymorphicSignature
    public final native Object invoke(Object... p0) throws Throwable;
    
    @MethodHandle.PolymorphicSignature
    final native Object invokeBasic(Object... p0) throws Throwable;
    
    @MethodHandle.PolymorphicSignature
    static native Object linkToVirtual(Object... p0) throws Throwable;
    
    @MethodHandle.PolymorphicSignature
    static native Object linkToStatic(Object... p0) throws Throwable;
    
    @MethodHandle.PolymorphicSignature
    static native Object linkToSpecial(Object... p0) throws Throwable;
    
    @MethodHandle.PolymorphicSignature
    static native Object linkToInterface(Object... p0) throws Throwable;
    
    public Object invokeWithArguments(Object... array) throws Throwable {
        MethodType genericMethodType = MethodType.genericMethodType((array == null) ? 0 : array.length);
        return genericMethodType.invokers().spreadInvoker(0).invokeExact(this.asType(genericMethodType), array);
    }
    
    public Object invokeWithArguments(List<?> list) throws Throwable {
        return this.invokeWithArguments(list.toArray());
    }
    
    public MethodHandle asType(MethodType methodType) {
        if (methodType == this.type) {
            return this;
        }
        MethodHandle typeCached = this.asTypeCached(methodType);
        if (typeCached != null) {
            return typeCached;
        }
        return this.asTypeUncached(methodType);
    }
    
    private MethodHandle asTypeCached(MethodType methodType) {
        MethodHandle asTypeCache = this.asTypeCache;
        if (asTypeCache != null && methodType == asTypeCache.type) {
            return asTypeCache;
        }
        return null;
    }
    
    MethodHandle asTypeUncached(MethodType obj) {
        if (!this.type.isConvertibleTo(obj)) {
            throw new WrongMethodTypeException("cannot convert " + this + " to " + obj);
        }
        return this.asTypeCache = MethodHandleImpl.makePairwiseConvert(this, obj, true);
    }
    
    public MethodHandle asSpreader(Class<?> clazz, int n) {
        MethodType spreaderChecks = this.asSpreaderChecks(clazz, n);
        int parameterCount = this.type().parameterCount();
        int n2 = parameterCount - n;
        BoundMethodHandle rebind = this.asType(spreaderChecks).rebind();
        return rebind.copyWith(spreaderChecks.replaceParameterTypes(n2, parameterCount, new Class[] { clazz }), rebind.editor().spreadArgumentsForm(1 + n2, clazz, n));
    }
    
    private MethodType asSpreaderChecks(Class<?> clazz, int n) {
        this.spreadArrayChecks(clazz, n);
        int parameterCount = this.type().parameterCount();
        if (parameterCount < n || n < 0) {
            throw MethodHandleStatics.newIllegalArgumentException("bad spread array length");
        }
        Class<Object> componentType = clazz.getComponentType();
        MethodType type = this.type();
        boolean n2 = true;
        boolean n3 = false;
        for (int i = parameterCount - n; i < parameterCount; ++i) {
            Class<Object> parameterType = type.parameterType(i);
            if (parameterType != componentType) {
                n2 = false;
                if (!MethodType.canConvert(componentType, parameterType)) {
                    n3 = true;
                    break;
                }
            }
        }
        if (!(n2 == 0)) {
            return type;
        }
        MethodType spreaderType = type.asSpreaderType(clazz, n);
        if (n3 == 0) {
            return spreaderType;
        }
        this.asType(spreaderType);
        throw MethodHandleStatics.newInternalError("should not return", null);
    }
    
    private void spreadArrayChecks(Class<?> clazz, int n) {
        Class<Object> componentType = clazz.getComponentType();
        if (componentType == null) {
            throw MethodHandleStatics.newIllegalArgumentException("not an array type", clazz);
        }
        if ((n & 0x7F) != n) {
            if ((n & 0xFF) != n) {
                throw MethodHandleStatics.newIllegalArgumentException("array length is not legal", Integer.valueOf(n));
            }
            assert n >= 128;
            if (componentType == Long.TYPE || componentType == Double.TYPE) {
                throw MethodHandleStatics.newIllegalArgumentException("array length is not legal for long[] or double[]", Integer.valueOf(n));
            }
        }
        return;
    }
    
    public MethodHandle asCollector(Class<?> clazz, int n) {
        this.asCollectorChecks(clazz, n);
        int n2 = this.type().parameterCount() - 1;
        BoundMethodHandle rebind = this.rebind();
        MethodType collectorType = this.type().asCollectorType(clazz, n);
        MethodHandle varargsArray = MethodHandleImpl.varargsArray(clazz, n);
        LambdaForm collectArgumentArrayForm = rebind.editor().collectArgumentArrayForm(1 + n2, varargsArray);
        if (collectArgumentArrayForm != null) {
            return rebind.copyWith(collectorType, collectArgumentArrayForm);
        }
        return rebind.copyWithExtendL(collectorType, rebind.editor().collectArgumentsForm(1 + n2, varargsArray.type().basicType()), varargsArray);
    }
    
    boolean asCollectorChecks(Class<?> clazz, int n) {
        this.spreadArrayChecks(clazz, n);
        int parameterCount = this.type().parameterCount();
        if (parameterCount != 0) {
            Class<Object> parameterType = this.type().parameterType(parameterCount - 1);
            if (parameterType == clazz) {
                return true;
            }
            if (parameterType.isAssignableFrom(clazz)) {
                return false;
            }
        }
        throw MethodHandleStatics.newIllegalArgumentException("array type not assignable to trailing argument", this, clazz);
    }
    
    public MethodHandle asVarargsCollector(Class<?> clazz) {
        clazz.getClass();
        int collectorChecks = this.asCollectorChecks(clazz, 0);
        if (this.isVarargsCollector() && collectorChecks) {
            return this;
        }
        return MethodHandleImpl.makeVarargsCollector(this, clazz);
    }
    
    public boolean isVarargsCollector() {
        return false;
    }
    
    public MethodHandle asFixedArity() {
        assert !this.isVarargsCollector();
        return this;
    }
    
    public MethodHandle bindTo(Object cast) {
        cast = this.type.leadingReferenceParameter().cast(cast);
        return this.bindArgumentL(0, cast);
    }
    
    @Override
    public String toString() {
        if (MethodHandleStatics.DEBUG_METHOD_HANDLE_NAMES) {
            return "MethodHandle" + this.debugString();
        }
        return this.standardString();
    }
    
    String standardString() {
        return "MethodHandle" + this.type;
    }
    
    String debugString() {
        return this.type + " : " + this.internalForm() + this.internalProperties();
    }
    
    BoundMethodHandle bindArgumentL(int n, Object o) {
        return this.rebind().bindArgumentL(n, o);
    }
    
    MethodHandle setVarargs(MemberName memberName) throws IllegalAccessException {
        if (!memberName.isVarargs()) {
            return this;
        }
        Class<Object> lastParameterType = this.type().lastParameterType();
        if (lastParameterType.isArray()) {
            return MethodHandleImpl.makeVarargsCollector(this, lastParameterType);
        }
        throw memberName.makeAccessException("cannot make variable arity", null);
    }
    
    MethodHandle viewAsType(MethodType methodType, boolean b) {
        assert !(!this.viewAsTypeChecks(methodType, b));
        BoundMethodHandle rebind = this.rebind();
        assert !(rebind instanceof DirectMethodHandle);
        return rebind.copyWith(methodType, rebind.form);
    }
    
    boolean viewAsTypeChecks(MethodType methodType, boolean b) {
        if (b) {
            assert !(!this.type().isViewableAs(methodType, true)) : Arrays.asList(new Object[] { this, methodType });
        }
        else
            assert !(!this.type().basicType().isViewableAs(methodType.basicType(), true)) : Arrays.asList(new Object[] { this, methodType });
        return true;
    }
    
    LambdaForm internalForm() {
        return this.form;
    }
    
    MemberName internalMemberName() {
        return null;
    }
    
    Class<?> internalCallerClass() {
        return null;
    }
    
    MethodHandleImpl.Intrinsic intrinsicName() {
        return MethodHandleImpl.Intrinsic.NONE;
    }
    
    MethodHandle withInternalMemberName(MemberName memberName, boolean b) {
        if (memberName != null) {
            return MethodHandleImpl.makeWrappedMember(this, memberName, b);
        }
        if (this.internalMemberName() == null) {
            return this;
        }
        BoundMethodHandle rebind = this.rebind();
        assert rebind.internalMemberName() == null;
        return rebind;
    }
    
    boolean isInvokeSpecial() {
        return false;
    }
    
    Object internalValues() {
        return null;
    }
    
    Object internalProperties() {
        return "";
    }
    
    abstract MethodHandle copyWith(MethodType p0, LambdaForm p1);
    
    abstract BoundMethodHandle rebind();
    
    void updateForm(LambdaForm x) {
        assert x.customized == this;
        if (this.form == x) {
            return;
        }
        x.prepare();
        MethodHandleStatics.UNSAFE.putObject(this, MethodHandle.FORM_OFFSET, x);
        MethodHandleStatics.UNSAFE.fullFence();
        return;
    }
    
    void customize() {
        if (this.form.customized == null) {
            this.updateForm(this.form.customize(this));
        }
        else
            assert this.form.customized == this;
    }
    
    static {
        MethodHandleImpl.initStatics();
        try {
            FORM_OFFSET = MethodHandleStatics.UNSAFE.objectFieldOffset(MethodHandle.class.getDeclaredField("form"));
        }
        catch (ReflectiveOperationException ex) {
            throw MethodHandleStatics.newInternalError(ex);
        }
    }
    
    @Target({ ElementType.METHOD })
    @Retention(RetentionPolicy.RUNTIME)
    @interface PolymorphicSignature {
    }
}
